from torch import Tensor
from torch.nn import LogSoftmax
import torch
from datetime import datetime as dt
from a_ukw_ml_tools.classes.prediction import MultilabelPrediction
from typing import List

SOFTMAX = LogSoftmax(dim=-1)

def mc_base_pred_dict(ai_config, timestamp = dt.now()):
    base_prediction_dict = {
        "name": ai_config.name,
        "version": ai_config.ai_settings.version,
        "date": timestamp,
        "value": None,
        "raw": None,
        "choices": ai_config.choices
    }
    return base_prediction_dict

def mc_batch_result_raw_to_pred_values(preds:Tensor, softmax = None):
    if not softmax: softmax=SOFTMAX
    raw = softmax(preds)
    values = torch.argmax(raw, dim = -1)

    if preds.is_cuda:
        values = values.cpu().numpy()
        raw = raw.cpu().numpy().tolist()
    else:
        values = values.numpy()
        raw = raw.numpy().tolist()

    return values, raw

def mc_prediction_to_object(value: int, raw_pred: List[int], base_pred_dict):
    prediction_dict = base_pred_dict.copy()
    prediction_dict["value"] = value
    prediction_dict["raw"] = raw_pred
    prediction = MultilabelPrediction(**prediction_dict)
    return prediction
    


import numpy as np
import pandas as pd
from ..mongodb.results import get_intervention_labels
from ..classes.prediction import BinaryPrediction
from datetime import datetime as dt

def smooth_binary_labels(preds, conv_len, weight_main = None, threshold = 0.5, future_frames = None):
    """
    Weight main of 0.5 would mean that the "current" frame is weighted 
    """
    if not conv_len % 2:
        conv_len += 1
    
    if not future_frames:
        future_frames = True

    if not weight_main:
        weight_main = 1/conv_len

    future_frames = False
    weight_main = conv_len * weight_main

    if future_frames:
        conv = np.ones(conv_len)
        
    else:
        conv = np.zeros(conv_len-1)
        conv = np.append(conv, np.ones(conv_len))

    mid = int(conv_len/2)+1
    conv[mid] = weight_main
    conv = conv/sum(conv)

    smooth_prediction = np.convolve(preds, conv, mode = "same")
    smooth_prediction[smooth_prediction>threshold] = True
    smooth_prediction[smooth_prediction<=threshold] = False
    return smooth_prediction

def calculate_intervention_smooth_labels(
    name,
    conv_len,
    intervention_id,
    version,
    db_images,
    future_frames = None,
    main_weight = None,
    threshold = 0.5,
    timestamp = None
    ):
    if not timestamp:
        timestamp = dt.now()
    _records = get_intervention_labels(intervention_id, db_images, "predictions")
    pred_df = pd.DataFrame.from_records(_records)

    select_df = pred_df[pred_df.name == name]
    select_df = select_df.sort_values("frame_number")
    preds = select_df.value.to_numpy()
    preds = smooth_binary_labels(preds, conv_len, main_weight, threshold, future_frames)
    for i, img_id in enumerate(select_df._id):
        value = preds[i]
        _pred = BinaryPrediction(**{
            "name": name,
            "version": version,
            "date": timestamp,
            "value": value,
            "raw": value,
            "choices": [False, True]}
        )
        db_images.update_one({"_id": img_id}, {"$set": {f"predictions_smooth.{name}": _pred.dict()}})





from typing import Collection, Iterable
from bson.objectid import ObjectId
from pymongo.collection import Collection
from tqdm import tqdm
from .ai_predictor import AiPredictor
from .fieldnames import FIELDNAME_VIDEO_KEY, FIELDNAME_VIDEO_PATH
from .terminology import Terminology
from ukw_ml_tools.db.crud import delete_frames_from_db
from .utils import *
from .fieldnames import *
import cv2
import pandas as pd

class Intervention:
    """asd
    asd
    """
    def __init__(self, identifier, db_images: Collection, db_interventions: Collection, cfg: dict, id_is_video_key: bool = False):
        self.db_images = db_images
        self.db_interventions = db_interventions
        self.cfg = cfg
        self.base_path_frames = Path(cfg["base_path_frames"])
        if id_is_video_key:
            self.intervention = self.db_interventions.find_one({FIELDNAME_VIDEO_KEY: identifier})
        else:
            self.intervention = self.db_interventions.find_one({"_id": identifier})
        
        assert self.intervention
        self._id = self.intervention["_id"]
        self.video_key = get_if_in_dict(self.intervention, [FIELDNAME_VIDEO_KEY])
        if self.video_key != None:
            self.frame_dir = self.base_path_frames.joinpath(self.intervention[FIELDNAME_VIDEO_KEY])
        else: self.frame_dir = None
        self.skip_frame_factor = cfg["skip_frame_factor"]
        # self.prediction_result: pd.DataFrame = None
        if not FIELDNAME_INTERVENTION_IMAGE_PREDICTIONS in self.intervention:
            self.update_image_predictions()


    def update_image_annotations(self) -> List[str]:
        labels = get_intervention_image_annotations(self._id, self.db_images)
        self.db_interventions.update_one({"_id": self._id}, {"$set": {FIELDNAME_INTERVENTION_IMAGE_ANNOTATIONS: labels}})
        self.refresh()

        return labels

    def update_image_predictions(self):
        predictions = get_intervention_image_predictions(self._id, self.db_images)
        self.db_interventions.update_one({"_id": self._id}, {"$set": {FIELDNAME_INTERVENTION_IMAGE_PREDICTIONS: predictions}})        
        self.refresh()
        return predictions

    def update_smooth_labels(self):
        assert FIELDNAME_VIDEO_METADATA in self.intervention
        assert FIELDNAME_FPS in self.intervention[FIELDNAME_VIDEO_METADATA]

        ####################################
        n_smooth = int(self.intervention[FIELDNAME_VIDEO_METADATA][FIELDNAME_FPS]/2)
        threshold = 0.5
        ####################################

        df = self.get_frame_result_df(n_smooth, threshold)
        df = df.loc[df[COLNAME_RESULT_TYPE]==RESULT_TYPE_PREDICTION_SMOOTH]
        ####################################
        df = df.loc[df[COLNAME_RESULT_NAME]!= "blurry"]
        ####################################

        for index, row in df.iterrows():
            self.db_images.update_one(
                {"_id": row["_id"]},
                {"$set": {f"{FIELDNAME_PREDICTIONS_SMOOTH}.{row[COLNAME_RESULT_NAME]}": row[COLNAME_RESULT_LABEL]}})

        self.refresh()

    def calculate_test_result(self, n_smooth, use_future, running_mean_treshold):
        self.get_frame_result_df(n_smooth, running_mean_treshold, use_future)
        

    def update_tokens(self):
        terminology = Terminology(self.cfg)

        report = get_if_in_dict(self.intervention, [FIELDNAME_REPORT_RAW])
        if report:
            tokens_report = terminology.get_terminology_result(report)
        else: 
            tokens_report = []
        patho = get_if_in_dict(self.intervention, [FIELDNAME_PATHO_RAW])
        if patho:
            tokens_patho = terminology.get_terminology_result(patho)
        else: 
            tokens_patho = []

        update = {
            FIELDNAME_TOKENS: {
                FIELDNAME_TOKENS_REPORT: tokens_report,
                FIELDNAME_TOKENS_PATHO: tokens_patho
            }
        }
        self.db_interventions.update_one({"_id": self._id}, {"$set": update})
        self.refresh()
        return update

    def get_video_metadata(self):
        return get_video_meta(Path(self.intervention[FIELDNAME_VIDEO_PATH]))

    def get_summary(self):
        flanks = get_if_in_dict(self.intervention, [FIELDNAME_PREDICTION_FLANKS])
        if flanks:
            flanks = {key: value[FIELDNAME_FLANK_RANGES] for key, value in flanks.items()}
        self.summary = {
            "_id": self.intervention["_id"],
            FIELDNAME_N_EXTRACTED_FRAMES: get_if_in_dict(self.intervention, [FIELDNAME_N_EXTRACTED_FRAMES]),
            FIELDNAME_FRAMES_TOTAL: get_if_in_dict(self.intervention, [FIELDNAME_FRAMES_TOTAL]),
            FIELDNAME_VIDEO_KEY: get_if_in_dict(self.intervention, [FIELDNAME_VIDEO_KEY]),
            FIELDNAME_VIDEO_PATH: get_if_in_dict(self.intervention, [FIELDNAME_VIDEO_PATH]),
            FIELDNAME_REPORT_RAW: get_if_in_dict(self.intervention, [FIELDNAME_REPORT_RAW]),
            FIELDNAME_PATHO_RAW: get_if_in_dict(self.intervention, [FIELDNAME_PATHO_RAW]),
            FIELDNAME_TOKENS: get_if_in_dict(self.intervention, [FIELDNAME_TOKENS]),
            FIELDNAME_INTERVENTION_IMAGE_PREDICTIONS: get_if_in_dict(self.intervention, [FIELDNAME_INTERVENTION_IMAGE_PREDICTIONS]),
            FIELDNAME_INTERVENTION_IMAGE_ANNOTATIONS: get_if_in_dict(self.intervention, [FIELDNAME_INTERVENTION_IMAGE_ANNOTATIONS]),
            FIELDNAME_PREDICTION_FLANKS: flanks
        }

        return self.summary

    def get_frame_id(self, n_frame: int):
        return self.intervention[FIELDNAME_FRAMES][str(n_frame)]

    def get_images_with_predictions(self, as_list: bool = False) -> Iterable:
        frame_ids = self.get_all_frame_ids()
        frames = self.db_images.find(
            {
                "_id": {
                    "$in": frame_ids
                },
                FIELDNAME_PREDICTIONS: {
                    "$nin": [{}]
                }
            }
        )

        if as_list: return [_ for _ in frames]
        else: return frames

    def get_images_with_annotations(self, as_list: bool = False):
        frame_ids = self.get_all_frame_ids()
        frames = self.db_images.find(
            {
                "_id": {
                    "$in": frame_ids
                },
                FIELDNAME_LABELS: {
                    "$nin": [{}]
                }
            }
        )
        if as_list: return [_ for _ in frames]
        else: return frames

    def get_all_frame_ids(self) -> List[ObjectId]:
        return [_id for n, _id in self.intervention[FIELDNAME_FRAMES].items()]

    def refresh(self):
        self.intervention = self.db_interventions.find_one({"_id": self._id})
        return self.intervention

    def get_prediction_result(self, as_list: bool = False) -> pd.DataFrame:
        frames_with_prediction = self.get_images_with_predictions()
        df_records = []

        for frame in frames_with_prediction:
            _predictions = frame[FIELDNAME_PREDICTIONS]
            predictions = [{
                "_id": frame["_id"],
                FIELDNAME_FRAME_NUMBER: frame[FIELDNAME_FRAME_NUMBER],
                COLNAME_RESULT_NAME: ai_name,
                COLNAME_RESULT_TYPE: RESULT_TYPE_PREDICTION,
                COLNAME_PREDICTION_VALUE: label_prediction[COLNAME_PREDICTION_VALUE],
                COLNAME_RESULT_LABEL: label_prediction[COLNAME_RESULT_LABEL],
                FIELDNAME_AI_VERSION: label_prediction[FIELDNAME_AI_VERSION] 
            } for ai_name, label_prediction in _predictions.items()]
            df_records.extend(predictions)

        if df_records:
            df = pd.DataFrame.from_records(df_records)
            df.sort_values([FIELDNAME_FRAME_NUMBER, COLNAME_PREDICTION_VALUE], ignore_index = True, inplace = True)
            self.prediction_df = df
        else:
            self.prediction_df = pd.DataFrame(
                columns = [
                    "_id",
                    FIELDNAME_FRAME_NUMBER,
                    COLNAME_RESULT_NAME,
                    COLNAME_RESULT_TYPE,
                    COLNAME_PREDICTION_VALUE,
                    COLNAME_RESULT_LABEL,
                    FIELDNAME_AI_VERSION                    
                ]
            ) 

        if as_list: return df_records
        else: return self.prediction_df

    def get_smooth_prediction_result(self, n_smooth: int = 10, threshold: float = 0.5, use_future: bool = True):
        names = self.prediction_df[COLNAME_RESULT_NAME].unique()

        for name in names:
            if name == "blurry":
                continue
            self.add_smooth_label_prediction(name, n_smooth, threshold, use_future)

    def get_annotation_result(self, as_list: bool = False) -> pd.DataFrame:
        frames = self.get_images_with_annotations()
        df_records = []

        for frame in frames:
            _labels = frame[FIELDNAME_LABELS]
            labels = [{
                "_id": frame["_id"],
                FIELDNAME_FRAME_NUMBER: frame[FIELDNAME_FRAME_NUMBER],
                COLNAME_RESULT_NAME: label,
                COLNAME_RESULT_TYPE: "annotation",
                COLNAME_RESULT_LABEL: choice,
            } for label, choice in _labels.items()]
            df_records.extend(labels)

        if df_records:
            df = pd.DataFrame.from_records(df_records)
            df.sort_values([FIELDNAME_FRAME_NUMBER], ignore_index = True,inplace=True)
            self.annotation_df = df
        else: 
            self.annotation_df = pd.DataFrame(columns = ["_id", FIELDNAME_FRAME_NUMBER, COLNAME_RESULT_NAME, COLNAME_RESULT_TYPE, COLNAME_RESULT_LABEL])

        if as_list: return df_records
        else: return self.annotation_df

    def get_frame_result_df(self, n_smooth: int = None, threshold: float = 0.5, use_future: bool = True):
        if n_smooth == None:
            n_smooth = int(self.intervention[FIELDNAME_VIDEO_METADATA][FIELDNAME_FPS]/1)
        self.get_annotation_result()
        self.get_prediction_result()
        self.get_smooth_prediction_result(n_smooth, threshold, use_future)
        self.frame_result_df = self.annotation_df.append(self.prediction_df)
        get_prediction_flanks(self)
        
        return self.frame_result_df

    def add_smooth_label_prediction(self, label, n_smooth: int = 10, threshold: float = 0.5, use_future: bool = False):
        smooth_prediction = get_smooth_prediction_df(self.prediction_df, label, n_smooth, threshold)
        self.prediction_df = self.prediction_df.append(smooth_prediction)

        return self.prediction_df

    def predict_all_frames(self, ai_name):
        frame_list = self.get_frame_list(as_list = True)

        predictor = AiPredictor(ai_name, self.cfg, self.db_images)
        predictor.base_paths_models = Path(self.cfg["base_path_models"])

        predictor.load_model()
        predictor.load_dataset(frame_list)
        predictor.load_dataloader()

        print(predictor.ai_name)
        print(predictor.specs)

        targets, predicted_batches = predictor.predict_images()

        for batch in predicted_batches:
            for _id, prediction in batch.items():
                self.db_images.update_one({"_id": _id}, {"$set": {f"{FIELDNAME_PREDICTIONS}.{ai_name}": prediction}})

        self.update_image_predictions()

    def prediction_df_long_to_wide(self, prediction_df: pd.DataFrame = None) -> pd.DataFrame:
        if not prediction_df:
            if self.prediction_df:
                prediction_df = self.prediction_df
            else:
                self.set_prediction_result()
                prediction_df = self.prediction_df

        return prediction_df_long_to_wide(prediction_df)
        
    def add_frame_labels(self, frame_labels: dict):
        for n_frame, _label in frame_labels.items():
            update_dict = self.frame_label_dict_to_update_dict(_label)
            self.db_images.update_one({"_id": self.get_frame_id(n_frame)}, {"$set": update_dict})

        self.refresh()

        return self.intervention

    def frame_label_dict_to_update_dict(self, frame_label_dict: dict):
        update_dict = {}
        for label, value in frame_label_dict.items():
            update_dict[f"{FIELDNAME_LABELS}.{label}"] = value

        return update_dict

    def get_frame_list(self, as_list: bool = True) -> List[dict]:
        frame_ids = [_id for n_frame, _id in self.intervention[FIELDNAME_FRAMES].items()]
        frame_cursor = self.db_images.find({"_id": {"$in": frame_ids}}) 
        if as_list:
            frame_list = [_ for _ in frame_cursor]
            return frame_list
        else:
            return frame_cursor

    def extract_frames(self, frame_list: List[int] = None, frame_suffix: str = ".jpg"):
        cap = cv2.VideoCapture(self.intervention[FIELDNAME_VIDEO_PATH])
        if not frame_list:
            frames_total = get_frames_total(cap)
            frame_list = [_ for _ in range(frames_total)]
            
        
        frame_list = [n_frame for n_frame in frame_list if str(n_frame) not in self.intervention[FIELDNAME_FRAMES]]
        frame_list.sort()
        # frame_dicts = []
        if not self.frame_dir.exists():
            os.mkdir(self.frame_dir)


        for n_frame in tqdm(frame_list):
            frame_path = self.frame_dir.joinpath(f"{n_frame}{frame_suffix}")
            

            cap.set(cv2.CAP_PROP_POS_FRAMES, n_frame)
            success, image = cap.read()
            assert success

            assert cv2.imwrite(frame_path.as_posix(), image)

            db_image_entry = get_image_db_template(
                origin=self.intervention[FIELDNAME_ORIGIN],
                intervention_id=self.intervention["_id"],
                path=frame_path.as_posix(),
                n_frame=n_frame,
                image_type= IMAGETYPE_FRAME
            )
            self.db_images.insert(db_image_entry)
            # frame_dicts.append(db_image_entry)

        cap.release()
        if frame_list:
            frames = self.db_images.find({FIELDNAME_INTERVENTION_ID: self._id})
            update = {
                FIELDNAME_FRAMES: {
                    str(frame[FIELDNAME_FRAME_NUMBER]): frame["_id"]
                    for frame in frames 
                    } 
                }

            self.db_interventions.update_one({"_id": self._id}, {"$set": update})
            self.intervention = self.refresh()

            return frames

        self.refresh()

    def delete_frames(self, frame_list: List[int]):
        frame_ids = [self.intervention[FIELDNAME_FRAMES][str(n_frame)] for n_frame in frame_list]
        frames = self.db_images.find({"_id": {"$in": frame_ids}})
        for frame in frames:
            os.remove(frame[FIELDNAME_IMAGE_PATH])
        self.db_images.delete_many({"_id": {"$in": frame_ids}})
        self.db_interventions.update_one(
            {
                "_id": self._id
            },
            {
                "$unset": {f"{FIELDNAME_FRAMES}.{n_frame}": "" for n_frame in frame_list}
            }              
        )

        self.refresh()




from a_ukw_ml_tools.mongodb.validation import check_intervention_frames_extracted
from a_ukw_ml_tools.mongodb.update import update_many_frame_is_extracted
from ..mongodb.get_objects import get_intervention

def get_video_keys_of_not_extracted_frames(db_images):
    """
    Matches metadata.is_extracted: False and metadata.path: {"$exists": True}
    """
    
    r = db_images.aggregate([
        {"$match": {"metadata.is_extracted": False, "metadata.path": {"$exists": True}}},
        {
            "$group": {
                "_id": "$video_key"
            }
        }
    ])
    return [_["_id"] for _ in r]

def update_frames_extracted(video_key, db_interventions, db_images):
    intervention = get_intervention(video_key, db_interventions)

    is_extracted_dict = check_intervention_frames_extracted(video_key, db_images, db_interventions)
    ids_extracted = [intervention.frames[i] for i, value in is_extracted_dict.items() if value]
    ids_not_extracted = [intervention.frames[i] for i, value in is_extracted_dict.items() if not value]

    update_many_frame_is_extracted(True, ids_extracted, db_images)
    update_many_frame_is_extracted(False, ids_not_extracted, db_images)


from .import_videos import import_new_extern_interventions
from .import_video_annotations import (
    get_new_annotation_videos,
    import_new_video_annotation,
    extract_and_generate_frames,
    set_image_label_updates_from_video_segmentations
)
from .check_extracted_frames import (
    get_video_keys_of_not_extracted_frames,
    update_frames_extracted
)
from ..classes.config import Configuration, AiLabelConfig
from datetime import datetime as dt
import time
from .set_compund_labels import set_compound_labels

def daily_service_pipeline(cfg:Configuration, ai_config_dict, ai_config:AiLabelConfig):
    db_images, db_interventions, db_test_data, db_train_data = cfg.get_databases()
    url, auth = cfg.get_extern_tuple()
    label_settings = cfg.label_settings
    base_dir_frames = cfg.base_paths.frames



    log = {}
    log["new_intervention_ids"] = import_new_extern_interventions(url, auth, db_interventions)
    
    annotated_new = get_new_annotation_videos(url, auth, db_interventions)

    # Import new flank annotations to videos
    print("Import new annotated videos")
    for annotated_video in annotated_new:
        r = import_new_video_annotation(annotated_video, label_settings, url, auth, db_interventions)


    # Extract Frames and Generate Frame Entries for DB
    print("Extract Frames for new annotations")
    log["extracted_frames"] = {}
    for annotated_video in annotated_new:
        log[annotated_video.video_key] = extract_and_generate_frames(
            annotated_video, base_dir_frames, url, auth, db_interventions, db_images
        )

    # Set Image Labels
    print("Set new annotations")
    for annotated_video in annotated_new:
        set_image_label_updates_from_video_segmentations(
            annotated_video,
            ai_config_dict,
            db_interventions,
            db_images,
            set_in_db = True
        )

    # Set Colo Segmentation Labels
    print("Calculate and set compound labels: Colo Segmentation")
    name = "colo_segmentation"
    exclude_initially = [
        "outside",
        "blurry"
    ]
    base_annotation_dict = {
        "source": "calculation",
        "annotator_id": 0,
        "date": dt.now(),
        "name": name,
        "choices": ai_config.choices,
    }
    set_compound_labels(name, base_annotation_dict, ai_config, db_images)

    
    

    time.sleep(60*10*60*1.5)
    video_keys = get_video_keys_of_not_extracted_frames(db_images)
    
    for video_key in video_keys:
        update_frames_extracted(video_key, db_interventions, db_images)

    return log


from pydantic import BaseModel, Field
from .base import PyObjectId
from pymongo.collection import Collection

class UpdateOne(BaseModel):
    collection: Collection
    match: dict
    operation: dict

    class Config:
        arbitrary_types_allowed = True

    def get_query(self):
        q = dict(self.match), {dict(self.operation)}
        return q

    def post(self):
        r = self.collection.update_one(self.get_query())


import albumentations as A
import cv2
import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision import transforms
# from .utils import crop_img

def crop_img(img):
    no_crop = False
    mask = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    hist = cv2.calcHist([img], [0], mask = None, histSize = [100], ranges = [0,256])
    n_max_bin = np.argmax(hist)
    threshold = 255/100 * (n_max_bin+1)
    if threshold <= 13:
        threshold = 13
    if threshold >= 100:
        threshold = 30

    mask[mask > threshold] = 255
    mask[mask <=threshold] = 0

    contours, hierachy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    if len(contours) == 1:
        contour = contours[0]
    elif len(contours) == 0:
        no_crop = True
    else:
        contour = max(contours, key = cv2.contourArea)

    if not no_crop:
        x,y,w,h = cv2.boundingRect(contour)
    else:
        x = 0
        y = 0
        _shape = img.shape
        
        h = _shape[0]
        w = _shape[1]
        # no_crop = False

    img = img[y:y+h, x:x+w]

    delta = w-h

    if delta < 0:
        # x axis needs to be expanded
        _padding = [(0,0), (abs(delta),0), (0,0)]
    else: 
        # y axis needs to be expanded
        _padding = [(delta,0), (0,0), (0,0)]

    img = np.pad(img, _padding)

    return img




img_augmentations = A.Compose(
    [
        A.HorizontalFlip(p=0.5),
        A.MotionBlur(p=0.3, blur_limit=(10, 20)),
        A.RandomBrightnessContrast(p=0.5),
    ]
)

cropping_large = A.Compose([A.CenterCrop(width=1024, height=1024)])

cropping_small = A.Compose([A.CenterCrop(width=720, height=720)])

img_transforms = transforms.Compose([transforms.ToTensor()])


class BinaryImageClassificationDS(Dataset):
    def __init__(self, paths, labels, scaling: int = 75, training: bool = True):
        self.paths = paths
        self.scaling = scaling
        self.training = training

        self.labels = labels
        assert len(paths) == len(labels)

        self.classes = {0: "negative", 1: "positive"}

    def __getitem__(self, idx):
        img = cv2.imread(self.paths[idx])
        width = int(1024 * self.scaling / 100)  # img.shape[1]
        height = int(1024 * self.scaling / 100)  # img.shape[0]

        # if img.shape[1] >= 1024 and img.shape[0] >= 1024:
        #     img = cropping_large(image=img)["image"]
        # elif img.shape[1] >= 720 and img.shape[0] >= 720:
        #     img = cropping_small(image=img)["image"]
        img = crop_img(img)
        dim = (width, height)

        # Switch BGR to RGB
        # img = np.flip(img, axis=-1)
        
        if self.training:
            img = img_augmentations(image=img)["image"]

        img = cv2.resize(img, dsize=dim, interpolation=cv2.INTER_AREA)
        img = cv2.normalize(
            img, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
        if self.training:
            return img_transforms(img), torch.tensor(self.labels[idx])
        else:
            return img_transforms(img), self.labels[idx]

    def __len__(self):
        return len(self.paths)



from tqdm import tqdm
from bson import ObjectId
from a_ukw_ml_tools.labels.multiclass_prediction import (
    mc_batch_result_raw_to_pred_values,
    mc_base_pred_dict,
    mc_prediction_to_object
)
from ..classes.intervention import Intervention
from ..mongodb.get_objects import get_images_by_id_list

def mc_prediction_and_upload(model, ai_config, dataloader, db_images):
    base_prediction_dict = mc_base_pred_dict(ai_config)

    for imgs, ids in tqdm(dataloader):
        preds = model.forward(imgs.to(0))
        values, raw = mc_batch_result_raw_to_pred_values(preds)
        for i, value in enumerate(values):
            _pred = mc_prediction_to_object(value, raw[i], base_prediction_dict)
            db_images.update_one(
                {"_id": ObjectId(ids[i])},
                {"$set":{f"predictions.{ai_config.name}": _pred.dict()}}
            )


def mc_to_binary_labels(name, intervention:Intervention, db_images):
    frame_df = intervention.frame_df()
    frames = get_images_by_id_list(frame_df.id.to_list(), db_images)

    for frame in frames:
        predictions = frame.predictions[name].to_binary_predictions()
        _update = {f"predictions.{prediction.name}": prediction.dict() for prediction in predictions}
        db_images.update_one({"_id": frame.id}, {"$set": _update})


from typing import Collection
from bson.objectid import ObjectId
from pymongo.collection import Collection
from .utils import delete_frame_from_intervention

class Image:
    """asd
    asd
    """
    def __init__(self, object_id: ObjectId, db_images: Collection, db_interventions: Collection):
        self.db_images = db_images
        self.db_interventions = db_interventions
        self.image = self.db_images.find_one({"_id": object_id})
        assert self.image
        self.n_frame = self.image["n_frame"]
        self._id = self.image["_id"]

    def delete(self):
        _ = delete_frame_from_intervention(
            self.image,
            self.db_interventions
        )

        self.db_images.delete_one({"_id": self._id})
        return True

    



from bson.objectid import ObjectId
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from typing import List
import warnings
from pathlib import Path
from .utils import *
from collections import defaultdict
import pandas as pd

from .fieldnames import *

aggregations = {}


class DbImages:
    """Add Documentation
    asd
    """

    def __init__(self, db_images: Collection, cfg: dict):
        self.db = db_images
        self.cfg = cfg
        self.exclude_origins_for_prelabeling = cfg["exclude_origins_for_prelabeling"]
        self.task_modes = cfg["tasks"]["modes"]
        

    # Queries
    ## Read
    def get_all(self, as_list: bool = False) -> (Cursor or List):
        images = self.db.find({})

        if as_list: return [_ for _ in images]
        else: return images

    def get_by_id(self, _id: ObjectId) -> dict:
        return self.db.find_one({"_id": _id})

    def get_by_id_list(self, _ids: List[ObjectId], as_list: bool=True) -> List[dict] or Cursor:
        images = self.db.find({"_id": {"$in": _ids}})

        if as_list: return [_ for _ in images]
        else: return images

    def get_train_images(
        self,
        label: str,
        value,
        test_intervention_ids: List[ObjectId],
        extend_conditions: List = None,
        extend_agg: List = None,
        as_list: bool = False
        ):

        aggregation = get_train_image_query(label, value, test_intervention_ids, extend_conditions, extend_agg)

        # conditions = [
        #     field_value_query(f"{FIELDNAME_LABELS}.{label}", value),
        #     fieldvalue_nin_list_query(FIELDNAME_INTERVENTION_ID, test_intervention_ids)
        # ]

        # if extend_conditions:
        #     conditions.extend(extend_conditions)

        # aggregation = [match_logical_and_aggregation(conditions)]

        # if extend_agg:
        #     aggregation.extend(extend_agg)

        try:
            images = self.db.aggregate(aggregation)
        except:
            print(aggregation)
            raise Exception
            return False
        if as_list:
            images = [_ for _ in images]

        return images


    def get_ls_task_images(
        self,
        label,
        upper_confidence_threshold: float = 50,
        lower_confidence_threshold: float = 50,
        mode: str = "high_conf",
        limit: int = 10000,
        exclude_origins: List[str] = None
    ):
        if not exclude_origins:
            if isinstance(exclude_origins, List):
                pass
            else: 
                exclude_origins = self.exclude_origins_for_prelabeling 

        assert mode in self.task_modes
        latest_ai_version = max(self.db.distinct(
            f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_AI_VERSION}"
        ))
        if mode == "high_conf":
            has_label = False
            confidence_aggregation = {
                "$or": [
                    {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$gt": upper_confidence_threshold}},
                    {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$lt": lower_confidence_threshold}},
                ]
            }
        elif mode == "low_conf":
            has_label = False
            confidence_aggregation = {
                "$and": [
                    {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$lt": upper_confidence_threshold}},
                    {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$gt": lower_confidence_threshold}},
                ]
            }
        elif mode == "contradicting":
            has_label = True
            confidence_aggregation = {
                "$or": [
                    {
                        "$and": [
                            {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$gt": upper_confidence_threshold}},
                            {f"{FIELDNAME_LABELS}.{label}": False},
                            {f"{FIELDNAME_LABELS_VALIDATED}.{label}": {"$exists": False}}
                        ]
                    },{
                        "$and": [
                            {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_PREDICTION_VALUE}": {"$lt": lower_confidence_threshold}},
                            {f"{FIELDNAME_LABELS}.{label}": True},
                            {f"{FIELDNAME_LABELS_VALIDATED}.{label}": {"$exists": False}}
                        ]
                    }
                ]
            }

        _images = self.db.aggregate(
            [
                {
                    "$match": {
                        "$and": [
                            {f"{FIELDNAME_LABELS}.{label}": {"$exists": has_label}},  # Unlabeled Images
                            {
                                "$or": [
                                    {
                                        f"{FIELDNAME_LABELS_UNCLEAR}.{LABEL_UNCLEAR}": False
                                    }, 
                                    {f"{FIELDNAME_LABELS_UNCLEAR}.{LABEL_UNCLEAR}": {"$exists": False}},
                                ]
                            },
                            {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_AI_VERSION}": latest_ai_version},
                            {FIELDNAME_ORIGIN: {"$nin": exclude_origins}},
                            confidence_aggregation,
                        ]
                    }
                },
                {"$sample": {"size": limit}},
            ]
        )

        _ids = []
        images = []
        for image in _images:
            if image["_id"] not in _ids:
                _ids.append(image["_id"])
                images.append(image)

        return images

    def get_prelabel_images(
        self,
        label: str,
        version: float = None,
        origin: str = None,
        limit: int = 1000,
        intervention_types: List[str] = ["Koloskopie", "Unknown"],
        predict_annotated: bool = False,
        exclude_origins: List[str] = None,
        exclude_interventions: List[ObjectId] = None,
        return_df: bool = False
    ) -> List:
        agg_content = []
        if not exclude_origins:
            if isinstance(exclude_origins, List):
                pass
            else: 
                exclude_origins = self.exclude_origins_for_prelabeling


        if origin:
            agg_content.append({"$match": {FIELDNAME_ORIGIN: origin}})
        
        q1 = logical_or_aggregation([
                {f"{FIELDNAME_PREDICTIONS}.{label}": {"$exists": False}},
                {f"{FIELDNAME_PREDICTIONS}.{label}.{FIELDNAME_AI_VERSION}": {"$lt": version}}
            ])

        agg_content = [
            q1,
            {f"{FIELDNAME_LABELS}.{label}": {"$exists": predict_annotated}},
            {FIELDNAME_ORIGIN: {"$nin": exclude_origins}},
            {FIELDNAME_INTERVENTION_TYPE: {"$in": intervention_types}}
            ]
        
        if exclude_interventions:
            agg_content.append({FIELDNAME_INTERVENTION_ID: {"$nin": exclude_interventions}})

        agg = [match_logical_and_aggregation(agg_content)]

        agg.append({"$sample": {"size": limit}})
        images = [_ for _ in self.db.aggregate(agg)]

        if return_df:
            df_dict = {"file_path": [], "_id": []}
            for _ in images:
                df_dict["_id"].append(_["_id"])
                df_dict["file_path"].append(_[FIELDNAME_IMAGE_PATH])
            images = pd.DataFrame().from_dict(df_dict)

        return images

    def get_label_count(
        self,
        label_list: List[str],
        exclude_intervention_ids: List[ObjectId] = []
    ) -> dict:
        count = defaultdict(dict)
        for label in label_list:
            r = self.db.aggregate(
                [
                    {
                        "$match": {
                            FIELDNAME_INTERVENTION_ID: {"$nin": exclude_intervention_ids}
                        }
                    },
                    {
                        "$group": {
                            "_id": f"{FIELDNAME_LABELS}.{label}",
                            "count": {"$sum": 1}
                        }
                    }
                ]
            )

            for category_count in r:
                if category_count["_id"] is None:
                    continue
                count[label][category_count["_id"]] = category_count["count"]

        return count

    def get_label_count_by_origin(
        self,
        label_list: List[str],
        exclude_intervention_ids: List[ObjectId]
        ):

        count = defaultdict(generate_default_dict)
        for label in label_list:
            r = self.db.aggregate(
                [
                    {
                        "$match": {
                            FIELDNAME_INTERVENTION_ID: {"$exists": True, "$nin": exclude_intervention_ids},
                            FIELDNAME_LABELS: {"$exists": True, "$nin": [{}]},
                            FIELDNAME_ORIGIN: {"$exists": True}
                        }
                    },
                    {
                        "$group": {
                            "_id": {FIELDNAME_LABELS: f"${FIELDNAME_LABELS}.{label}", FIELDNAME_ORIGIN: f"${FIELDNAME_ORIGIN}"},
                            "count": {"$sum": 1},
                        }
                    }
                ]
            )

            for category_count in r:
                count[label][category_count["_id"][FIELDNAME_ORIGIN]][
                    category_count["_id"][FIELDNAME_LABELS]
                ] = category_count["count"]

        return count

    def get_labelcount_by_intervention(self, label_list: List[str] = []):
        count = defaultdict(generate_default_dict)
        for label in label_list:
            r = self.db.aggregate(
                [
                    {
                        "$match": {
                            FIELDNAME_LABELS: {"$exists": True, "$nin": [{}]},
                            FIELDNAME_INTERVENTION_ID: {"$exists": True}
                        }
                    },
                    {
                        "$group": {
                            "_id": {
                                FIELDNAME_LABELS: f"${FIELDNAME_LABELS}.{label}",
                                FIELDNAME_INTERVENTION_ID: f"${FIELDNAME_INTERVENTION_ID}",
                            },
                            "count": {"$sum": 1},
                        }
                    }
                ]
            )

            for category_count in r:
                count[label][category_count["_id"][FIELDNAME_INTERVENTION_ID]][
                    category_count["_id"][FIELDNAME_LABELS]
                ] = category_count["count"]

        return count

    ## Update
    def update_image_predictions(self, label: str, update: List[dict]):
        for _id, _update in update.items():
            self.db.update_one({"_id": _id}, {"$set": {f"{FIELDNAME_PREDICTIONS}.{label}":_update}})


    def calculate_stats(self, return_records: bool = True):
        self.stats_queries = {
            "entity": PREFIX_IMAGE,
            "get_count": [],
            "get_grouped_count": [],
            "get_grouped_dict_count": [
                {
                    # All Labels Count
                    "fieldname": FIELDNAME_LABELS,
                    "additional_match_conditions": None,
                    "additional_group_conditions": None
                },
                {
                    # Labels by Origin
                    "fieldname": FIELDNAME_LABELS,
                    "additional_match_conditions": None,
                    "additional_group_conditions": {FIELDNAME_ORIGIN: "$"+FIELDNAME_ORIGIN}
                },
                # {
                #     # Labels by Intervention_id
                #     "fieldname": FIELDNAME_LABELS,
                #     "additional_match_conditions": None,
                #     "additional_group_conditions": {FIELDNAME_INTERVENTION_ID: "$"+FIELDNAME_INTERVENTION_ID}
                # },
            ]
        }

        stats = calculate_stats_dict(self.db, self.stats_queries, return_records = return_records)
        return stats


    # Stats
    # n_frames
    # n_frames with annotations
    # n_frames with predictions
    # n_frames with annotation + Prediction
    # n_frames by origin
    # n_frames by intervention
    # n_frames with label / value
    # n_frames with prediction / value
    # n_frames by label: annotation and prediction agree
    # n_frames by label: annotation and prediction contradicting 


    # Validation
    def validate_image_paths(self) -> List:
        images = self.get_all()
        no_image = [str(_["_id"]) for _ in images if not Path(_[FIELDNAME_IMAGE_PATH]).exists()]

        if no_image:
            warnings.warn("Images in Db without existing file were found")
            warnings.warn("\n".join(no_image))

        return no_image

    def validate_all_intervention_ids_exist(self) -> List:
        return []
    #     _ids = self.db.distinct(FIELDNAME_INTERVENTION_ID)
    #     intervention_not_found = []
    #     for _ in _ids:
    #         intervention = self.db_interventions.find_one({"_id": _})
    #         if not intervention:
    #             intervention_not_found.append(str(_))

    #     return intervention_not_found